"""
Copyright 2016 Farzad Ghanei

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

 http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import absolute_import, division, print_function

__title__ = "navdoon"
__summary__ = "Powerful Statsd server, made easy"
__version__ = "0.1.0"
__author__ = "Farzad Ghanei"
__license__ = "Apache License, Version 2.0"
__copyright__ = "Copyright 2016 Farzad Ghanei"
